# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import hashlib

import requests
from zxcvbn import zxcvbn
from typing import List, Tuple

import pass_import


class PwnedAPI():
    """Simple wrapper for https://haveibeenpwned.com API."""

    def __init__(self):
        self.headers = {
            'user-agent': f"{pass_import.__title__}/{pass_import.__version__}"}

    def password_range(self, prefix: str) -> Tuple[List[str], List[int]]:
        """Query the haveibeenpwned api to retrieve the bucket ``prefix``."""
        url = f"https://api.pwnedpasswords.com/range/{prefix}"
        res = requests.get(url, headers=self.headers, verify=True, timeout=5)
        res.raise_for_status()

        hashes = []
        counts = []
        for item in res.text.split('\r\n'):
            partialhash, count = item.split(':')
            hashes.append(prefix + partialhash)
            counts.append(int(count))
        return (hashes, counts)


class Audit():
    """Audit passwords for vulnerabilities.

    Based on the PassAudit class from pass-audit.
    See https://github.com/roddhjav/pass-audit for more information.

    :param list[dict] data: The list of password entries to audit
        Each password entry is a dictionary.

    """

    def __init__(self, data):
        self.data = data
        self.breached = []
        self.weak = []
        self.duplicated = []

    @property
    def report(self):
        """Get audit result."""
        return {
            'breached': self.breached,
            'weak': self.weak,
            'duplicated': self.duplicated
        }

    def password(self):
        """K-anonimity password breach detection on haveibeenpwned.com."""
        # Generate the list of hashes and prefixes to query.
        data = []
        api = PwnedAPI()
        buckets = {}
        for entry in self.data:
            if entry.get('password', '') == '':
                continue
            password = entry['password'].encode("utf8")
            phash = hashlib.sha1(password).hexdigest().upper()  # nosec
            prefix = phash[0:5]
            data.append((entry, phash, prefix))
            if prefix not in buckets:
                buckets[prefix] = api.password_range(prefix)

        # Compare the data and return the breached passwords.
        for entry, phash, prefix in data:
            if phash in buckets[prefix][0]:
                index = buckets[prefix][0].index(phash)
                count = buckets[prefix][1][index]
                self.breached.append((entry.get('password', ''), count))

    def zxcvbn(self):
        """Password strength estimaton usuing Dropbox' zxcvbn."""
        for entry in self.data:
            if entry.get('password', '') == '':
                continue
            password = entry['password']
            user_input = list(entry.values())
            if password in user_input:
                user_input.remove(password)
            results = zxcvbn(password, user_inputs=user_input)
            if results['score'] <= 2:
                self.weak.append((password, results))

    def duplicates(self):
        """Check for duplicated passwords."""
        seen = {}
        for entry in self.data:
            if entry.get('password', '') == '':
                continue
            password = entry['password']
            if password in seen:
                seen[password].append(entry)
            else:
                seen[password] = [entry]

        for entries in seen.values():
            if len(entries) > 1:
                self.duplicated.append(entries)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import os
from typing import Dict
from abc import abstractmethod

from pass_import import clean
from pass_import.audit import Audit
from pass_import.core import Asset, Cap


class PasswordManager(Asset):
    """Interface for all password managers.

    **Manager metadata**

    :param str url: Public website of the password manager.
    :param str hexport: How to export data from the password manager.
    :param str himport: How to import data from the password manager.
    :param bool secure: A flag, to set to ``False`` if the password manager is
        considered not secure.

    **Set by reading settings**

    :param Action action: The current action for what the object is used.
    :param str root: Internal root where to import the passwords inside the pm.
    :param str delimiter: CSV delimiter character. Default: ``,``
    :param str cols: String that shows the list of CSV expected
        columns to map columns to credential attributes. Only used for the CSV
        generic importer.

    """
    url = ''
    hexport = ''
    himport = ''
    secure = True
    keys: Dict[str, str] = {}
    keyslist = [
        'title', 'password', 'login', 'email', 'url', 'comments', 'otpauth',
        'group'
    ]

    def __init__(self, prefix=None, settings=None):
        settings = {} if settings is None else settings

        self.data: Dict[str, str] = []
        self.root = settings.get('root', '')
        self.cols = settings.get('cols', '')
        self.action = settings.get('action', Cap.IMPORT)
        self.delimiter = str(settings.get('delimiter', ','))
        super().__init__(prefix)

    @classmethod
    def usage(cls) -> str:
        """Get password manager usage."""
        res = '\n'.join(cls.__doc__.split('\n')[1:-1])
        if ':usage:' in res:
            res = res.split(':usage:')[1]
            while '  ' in res:
                res = res.replace('  ', ' ')
            return res
        return ''

    @classmethod
    def description(cls) -> str:
        """Get password manager description."""
        return cls.__doc__.split('\n', maxsplit=1)[0]


class PasswordImporter(PasswordManager):
    """Interface for all password managers that support importing passwords.

    :param list[dict] data: The list of password entries imported by the parse
        method. Each password entry is a dictionary.
    :param list keyslist: The list of core key that will be present into the
        password entry, even without the extra option.
    :param dict keys: Correspondence dictionary between the password-store key
        name (``password``, ``title``, ``login``...), and the key name from the
        password manager considered.

    """
    cap = Cap.IMPORT

    @abstractmethod
    def parse(self):
        """Parse the password manager repository and retrieve passwords."""

    def invkeys(self) -> Dict[str, str]:
        """Return the invert of ``keys``."""
        return {v: k for k, v in self.keys.items()}

    def _sortgroup(self, folders: Dict[str, Dict[str, str]]):
        """Order groups in ``data``.

        :param dict folders: The group structure, it must be generated
            as follow:
                folders['<group-id>'] = {
                    'group': '<name>',
                    'parent': '<parent-id>'
                }
        """
        for folder in folders.values():
            parentid = folder.get('parent', '')
            parentname = folders.get(parentid, {}).get('group', '')
            folder['group'] = os.path.join(parentname, folder.get('group', ''))

        for entry in self.data:
            groupid = entry.get('group', '')
            entry['group'] = folders.get(groupid, {}).get('group', '')


class PasswordExporter(PasswordManager):
    """Interface for all password managers that support exporting passwords.

    **Set by reading settings**

    :param bool all: Ethier or not import all the data. Default: ``False``
    :param bool force: Either or not to force the insert if the path already
        exist. Default: ``False``

    """
    cap = Cap.EXPORT

    def __init__(self, prefix=None, settings=None):
        settings = {} if settings is None else settings
        self.all = settings.get('all', False)
        self.force = settings.get('force', False)
        super().__init__(prefix, settings)

    @abstractmethod
    def insert(self, entry: Dict[str, str]):
        """Insert a password entry into the password repository.

        :param dict entry: The password entry to insert.
        :raises PMError: If the entry already exists or in case of
            a password manager error.
        """

    def clean(self, cmdclean: bool, convert: bool):
        """Clean data before export.

        **Features:**

        1. Remove unused keys and empty values.
        2. Clean the protocol's name in the title.
        3. Clean group from unwanted values in Unix or Windows paths.
        4. Duplicate paths.
        5. Format the One-Time Password (OTP) url.

        :param bool cmdclean:
            If ``True``, make the paths more command line friendly.
        :param bool convert:
            If ``True``, convert the invalid characters present in the paths.

        """
        for entry in self.data:
            entry = clean.unused(entry)
            path = clean.group(clean.protocol(entry.pop('group', '')))
            entry['path'] = clean.cpath(entry, path, cmdclean, convert)

        clean.dpaths(self.data, cmdclean, convert)
        clean.dpaths(self.data, cmdclean, convert)
        clean.duplicate(self.data)
        clean.otp(self.data)

    def audit(self, hibp: bool = False):
        """Audit the parsed password for vulnerable passwords.

        **Features:**

        1. Look for breached password from haveibeenpwned.com,
        2. Check for duplicated passwords,
        3. Check password strength estimaton using zxcvbn.

        :param bool hibp: A flag, to set to ``True`` to look for breached
            password from haveibeenpwned.com
        :returns dict: A report dict.

        """
        audit = Audit(self.data)
        if hibp:
            audit.password()
        audit.zxcvbn()
        audit.duplicates()
        return audit.report



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2022 Artyom Yurash <urashav@ya.ru>.
#

from pass_import.core import register_managers
from pass_import.formats.csv import CSV


class SafeInCloudCSV(CSV):
    """Importer for SafeInCloud in CSV format."""
    name = 'safeincloud'
    url = 'https://safeincloud.ladesk.com/'
    hexport = 'File > Export > Comma-Separated Values (CSV)'
    himport = 'pass import safeincloud file.csv'
    keys = {
        'title': 'Title',
        'login': 'Login',
        'password': 'Password',
        'url': 'URL',
        'comments': 'Notes',
        'otpauth': 'OTP'
    }


register_managers(SafeInCloudCSV)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2022 Shashwat Pragya <shashwat183@gmail.com>.
#

from pass_import.formats.csv import CSV
from pass_import.core import register_managers


class NordPassCSV(CSV):
    """Importer for Nord Pass in CSV format."""
    name = 'nordpass'
    url = 'https://nordpass.com/'
    hexport = 'Settings > Export Items'
    himport = 'pass import nordpass file.csv'
    keys = {
        'title': 'name',
        'password': 'password',
        'login': 'username',
        'url': 'url',
        'comments': 'note',
        'group': 'folder'
    }

    def parse(self):
        """Parse NordPass CSV file."""
        super().parse()
        # NordPass exports individual folders as their own
        # empty rows. This code removes the extra folder entries
        # from the parsed data.
        groups = []
        for entry in self.data:
            if entry['group']:
                groups.append(entry['group'])

        self.data = list(filter(
            lambda x: not (x['title'] in groups and not x['group']),
            self.data))


register_managers(NordPassCSV)



#!/usr/bin/env python3
# -*- coding: utf-8 -*
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
# SPDX-License-Identifier: GPL-3.0-or-later

import io
import os
import sys
import traceback
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from pass_import import Detecters, Managers, __version__
from pass_import.auto import AutoDetect
from pass_import.core import Cap
from pass_import.errors import FormatError, PMError
from pass_import.tools import Config, get_magics

MANAGERS = Managers()

try:
    import jsonpath_ng.ext
    from jsonpath_ng.exceptions import JsonPathLexerError, JsonPathParserError
    JSONNG = True
except ImportError:
    JSONNG = False


class ArgParser(ArgumentParser):
    """Manages argument parsing and adds some defaults."""

    def __init__(self, passwordstore=False):
        self.passwordstore = passwordstore
        if self.passwordstore:
            prog = 'pass import'
            description = """
  Import data from most of the password manager. Passwords are imported into
  the existing default password store; therefore, the password store must have
  been initialised before with 'pass init'."""
        else:
            prog = 'pimport'
            description = """
  Import data from most of the password manager. Passwords are imported into
  an existing password repository; therefore, the password repository must have
  been initialised before."""

        super().__init__(
            prog=prog,
            description=description,
            formatter_class=RawDescriptionHelpFormatter,
            epilog="More information may be found in the "
                   f"{prog.replace(' ', '-')}(1) man page.",
            add_help=False)
        self.add_arguments()

    def add_arguments(self):
        """Set arguments for `pass import` or `pimport`."""
        pmarg = self.add_argument_group(title='Password managers')
        if not self.passwordstore:
            pmarg.add_argument(
                'dst', type=str, nargs='?', default='',
                help=("Destination password manager, can be: "
                      f"{', '.join(MANAGERS.names(Cap.EXPORT))}."))

        pmarg.add_argument(
            'src', type=str, nargs='*', default=[],
            help='Path to the data to import. Can also be the password manager'
                 ' name followed by the path to the data to import. The passw'
                 f"ord manager name can be: {', '.join(MANAGERS.names())}.")

        if not self.passwordstore:
            pmarg.add_argument(
                '-o', '--out', action='store', default='',
                help='Where the destination password manager lives. '
                     'Can be a file, a directory or even a login depending '
                     'of the manager.')

        # Common options
        common = self.add_argument_group(title='Common optional arguments')
        common.add_argument(
            '-r', '--root', action='store', dest='sroot',
            default='', metavar='path',
            help='Only import the password from a specific subfolder.')
        common.add_argument(
            '-p', '--path', action='store', dest='droot',
            default='', metavar='path',
            help='Import the passwords to a specific subfolder.')
        common.add_argument('-k', '--key', action='store', default='',
                            help='Path to a keyfile if required by a manager.')
        common.add_argument('-a', '--all', action='store_true',
                            help='Also import all the extra data present.')
        common.add_argument('-f', '--force', action='store_true',
                            help='Overwrite existing passwords.')
        common.add_argument('-c', '--clean', action='store_true',
                            help='Make the paths more command line friendly.')
        common.add_argument(
            '-C', '--convert', action='store_true',
            help='Convert invalid characters present in the paths.')
        common.add_argument(
            '-P', '--pwned', action='store_true',
            help='Check imported passwords against haveibeenpwned.com.')
        common.add_argument(
            '-d', '--dry-run', action='store_true',
            help='Do not import passwords, only show what would be imported.')

        # Extra options
        extra = self.add_argument_group(title='Extra optional arguments')
        extra.add_argument(
            '--sep', dest='separator', metavar='CHAR', default='-',
            help="Provide a characters of replacement for the path separator. "
                 "Default: '-'")
        extra.add_argument(
            '--del', dest='delimiter', metavar='CHAR', default=',',
            help="Provide an alternative CSV delimiter character. "
                 "Default: ','")
        extra.add_argument(
            '--cols', action='store', default='',
            help='CSV expected columns to map columns to credential attributes'
                 '. Only used by the csv importer.')
        extra.add_argument(
            '--filter', dest='filter', metavar='FILTER', default=None,
            help="""Export whole entries matching a JSONPath filter expression.
                 Default: (none)
                 This field can be:
                  - a string JSONPath expression
                  - an absolute path to a file containing a JSONPath filter
                     expression.
                 List of supported filter:
                 https://github.com/h2non/jsonpath-ng
                 Example:
                  - '$.entries[*].tags[?@="Defaults"]' : Export only entries
                 with a tag matching 'Defaults'""")
        extra.add_argument('--config', action='store', default='',
                           help="Set a config file. Default: '.import'")

        # Managers list
        usage = self.add_argument_group(
            title='Help related optional arguments')
        if self.passwordstore:
            usage.add_argument('-l', '--list', action='store_true',
                               help='List the supported password managers.')
        else:
            usage.add_argument('-l', '--list-importers', action='store_true',
                               help='List the supported password importers.')
            usage.add_argument('-e', '--list-exporters', action='store_true',
                               help='List the supported password exporters.')

        # Help and version
        usage.add_argument('-h', '--help', action='store_true',
                           help='Show this help message and exit.')
        usage.add_argument('-V', '--version', action='version',
                           version='%(prog)s ' + __version__,
                           help='Show the program version and exit.')
        group = usage.add_mutually_exclusive_group()
        group.add_argument('-v', '--verbose', action='count', default=0,
                           help='Set verbosity level, '
                                'can be used more than once.')
        group.add_argument('-q', '--quiet', action='store_true',
                           help='Be quiet.')

    def parse_args(self, args=None, namespace=None):
        """Parse pass-import arguments & print help."""
        if args is None:
            sys.argv.pop(0)
            args = sys.argv

        arg = vars(super().parse_args(args, namespace))
        arg['prog'] = self.prog
        if arg['help']:
            name = ''
            if self.passwordstore:
                if arg['src']:
                    name = arg['src'][0]
            else:
                name = arg['dst']

            if name in MANAGERS.names():
                self.print_help_manager(name)
            else:
                self.print_help()
            sys.exit(0)
        return arg

    def print_help_manager(self, name):
        """Print manager usage."""
        print(f'Usage: {self.prog} {name} [options]\n')
        for pm in MANAGERS.matrix().get(name):
            print(f'{pm.description()}:')
            usage = pm.usage()
            if usage:
                print(usage)
            if pm.format != '':
                print(f'  Format: {pm.format}')
            if pm.version != '':
                print(f'  Version: {pm.version}')
            if pm.url != '':
                print(f"  Url: {pm.url}")
            if pm.hexport != '':
                print(f'  Export: {pm.hexport}')
            if pm.himport != '':
                print(f'  Import: {pm.himport}')
            if pm.default:
                print(f'  This is the default importer for {name}.')
            if pm.cap is Cap.IMPORT | Cap.EXPORT:
                print('  Can be used for password import and export.')
            print()


def setup():
    """Read progam arguments, configuration & sanity checks."""
    conf = Config()
    parser = ArgParser(conf.passwordstore)
    arg = parser.parse_args()
    conf.verbosity(arg['verbose'], arg['quiet'])
    try:
        conf.readconfig(arg)
    except AttributeError as error:
        conf.verbose(error)
        conf.die("configuration file not valid.")
    conf.currate()

    if conf['list_importers'] or conf['list_exporters']:
        listmanagers(conf)

    if conf['exporter'] == '':
        conf.die("destination password manager not present.")

    if conf['exporter'] not in MANAGERS.names(Cap.EXPORT):
        conf.die(f"{conf['exporter']} is not a supported "
                 "destination password manager.")

    if not conf['src']:
        conf.die("The source password manager or the path to import is empty.")

    return conf


def listmanagers(conf):
    """List the supported password managers."""
    cap = Cap.IMPORT if conf['list_importers'] is True else Cap.EXPORT
    if conf.quiet:
        print('\n'.join(MANAGERS.names(cap)))
        sys.exit(0)

    if cap is Cap.EXPORT:
        msg = (f"The {len(MANAGERS.names(cap))} supported exporter "
               "password managers are:")
    else:
        msg = f"The {len(MANAGERS)} supported password managers are:"
    conf.success(msg)

    max_res = ''
    listing = {}
    matrix = MANAGERS.matrix(cap)
    for name in matrix:
        frmts = []
        for pm in matrix[name]:
            res = pm.format
            if pm.version:
                res += f' (v{pm.version})'
            max_res = max(max_res, res)
            frmts.append(res)
        listing[name] = frmts

    padding1 = len(max(MANAGERS.names(cap), key=len)) + 1
    if conf.verb:
        padding2 = len(max_res) + 1
        for name in sorted(matrix):
            for pm, frmt in zip(matrix[name], listing[name]):
                conf.message(conf.BOLD + name.ljust(padding1) + conf.end +
                             frmt.ljust(padding2) + pm.__name__)
    else:
        tmp = [', '.join(frmts) for frmts in listing.values()]
        padding2 = len(max(tmp, key=len)) + 1
        for name in sorted(listing):
            conf.message(conf.BOLD + name.ljust(padding1) + conf.end +
                         ', '.join(listing[name]).ljust(padding2) +
                         matrix[name][0].url)
    sys.exit(0)


def decryptsource(conf):
    """Decrypt source file if required."""
    path = conf['src'][1] if len(conf['src']) >= 2 else conf['src'][0]
    if os.path.isfile(path):
        decrypters = Detecters(Cap.DECRYPT)
        frmt, encoding = get_magics(path)
        if encoding:
            conf['encoding'] = encoding
        if frmt in decrypters:
            with decrypters[frmt](path) as file:
                conf['plaintext'] = file.decrypt()
                conf['decrypted'] = True
            conf.verbose(f"Source file decrypted using {frmt}.")


def detectmanager(conf):
    """Detect file format and password manager."""
    prefix = ''
    if len(conf['src']) == 1:
        name = conf['src'][0]
        if name in MANAGERS.names():
            conf.verbose("Using default manager.")
            detect = AutoDetect(name)
            pm = detect.default()

        else:
            conf.verbose("Trying to guess file format and manager name.")
            prefix = to_detect = name
            if conf['decrypted']:
                to_detect = conf['plaintext']

            detect = AutoDetect(settings=conf.getsettings())
            pm = detect.manager(to_detect)
            if pm is None:
                conf.die("Unable to detect the manager. Please try with: "
                         f"{conf['prog']} <manager> {prefix}")

    else:
        name = conf['src'][0]
        prefix = conf['src'][1]
        if name in MANAGERS.names():
            conf.verbose("Trying to guess file format.")
            to_detect = prefix
            if conf['decrypted']:
                to_detect = conf['plaintext']

            detect = AutoDetect(name, settings=conf.getsettings())
            pm = detect.format(to_detect)

        elif name in MANAGERS.clsnames():
            pm = MANAGERS.get(name)
            conf.verbose(f"Using import class: {pm.__name__}.")

        else:
            conf.die(f"{name} is not a supported source password manager.")

    conf.verbose(f"Importer: {pm.name}, Format: {pm.format}, Version: "
                 f" {pm.version}")

    if 'plaintext' in conf:
        conf['in'] = io.StringIO(conf['plaintext'])
    else:
        conf['in'] = prefix
    conf['importer'] = pm.name
    return pm


def zxcvbn_parse(details):
    """Nicely print the results from zxcvbn."""
    sequence = ''
    for seq in details.get('sequence', []):
        sequence += f"{seq['token']}({seq['pattern']}) "
    res = f"Score {details['score']} ({details['guesses']} guesses). "
    return res + f"This estimate is based on the sequence {sequence}"


# pylint: disable=inconsistent-return-statements
def pass_import(conf, cls_import):
    """Import data."""
    try:
        settings = conf.getsettings(conf['sroot'])
        with cls_import(conf['in'], settings=settings) as importer:
            importer.parse()
            if not importer.secure:  # pragma: no cover
                conf.warning(f"The password manager {conf['importer']} has "
                             "been flagged as unsecure, you should update all "
                             "your newly imported credentials.")
            return importer.data

    except (FormatError, AttributeError, ValueError, TypeError) as error:
        conf.debug(traceback.format_exc())
        conf.warning(error)
        conf.die(
            f"{conf['in']} is not a valid exported {conf['importer']} file.")

    except ImportError as error:
        conf.verbose(error)
        err = (f"Importing {conf['importer']}, missing required dependency: "
               f"{error.name}\n"
               f"You can install it with:\n  'pip3 install {error.name}'")
        if error.name not in ['pykeepass']:
            err += f", or\n  'sudo apt-get install python3-{error.name}'"
        conf.die(err)

    except (PermissionError, PMError) as error:
        conf.debug(traceback.format_exc())
        conf.die(error)


def pass_export(conf, cls_export, data):
    """Insert cleaned data into the password repository."""
    paths_imported = []
    paths_exported = []
    try:
        settings = conf.getsettings(conf['droot'], Cap.EXPORT)
        with cls_export(conf['out'], settings=settings) as exporter:
            exporter.data = data
            exporter.clean(conf['clean'], conf['convert'])
            report = exporter.audit(conf['pwned'])
            for entry in exporter.data:
                pmpath = os.path.join(conf['droot'], entry.get(
                    'path', entry.get('title', '')))
                conf.show(entry)
                exported = pass_filter(conf, entry)
                try:
                    if exported:
                        if not conf['dry_run']:
                            exporter.insert(entry)
                except PMError as error:
                    conf.debug(traceback.format_exc())
                    conf.warning(f"Impossible to insert {pmpath} into "
                                 f"{conf['exporter']}: {error}")
                else:
                    paths_imported.append(pmpath)
                    if exported:
                        paths_exported.append(pmpath)
    except PMError as error:
        conf.debug(traceback.format_exc())
        conf.die(error)

    return paths_imported, paths_exported, report


def pass_filter(conf, entry):
    """Filter entry based on a JSONPath filter expression."""
    filter_expression = conf.get('filter', None)
    if filter_expression is None:
        return True

    if not JSONNG:
        message = ("--filter requires pass-import[filter] "
                   "or pass-import[all] to be installed")
        raise ImportError('Missing packages. ' + message)

    # Having end users write their JSONPath filter expression as if
    # pass-import processes/filters entries in bulk, will allow end
    # users filter expression to continue to work when/if bulk filter is
    # supported. Hence 'entries' being added
    data = {'entries': [entry]}
    try:
        expr = jsonpath_ng.ext.parse(filter_expression)
        matches = expr.find(data)
        return len(matches) > 0
    except (JsonPathLexerError, JsonPathParserError) as e:
        conf.warning("Entry not exported due to error "
                     "most likely in FILTER expression"
                     + e.args[0])
        if conf.verb >= 2:
            conf.verbose("- Filter expression:" + filter_expression)
            conf.verbose("- Entry: " + data)
        return False


def report(conf, paths_imported, paths_exported, audit):
    """Print final success report."""
    if conf['dry_run']:
        conf.warning(f"Data would be imported from {conf['importer']} "
                     f"to {conf['exporter']}")
    else:
        conf.success(f"Importing passwords from {conf['importer']} "
                     f"to {conf['exporter']}")
    conf.message(f"Passwords imported from: {conf['in']}")
    conf.message(f"Passwords exported to: {conf['out']}")
    if conf['sroot'] != '':
        conf.message(f"Root path: {conf['sroot']}")
    if conf['droot'] != '':
        conf.message(f"Root path: {conf['droot']}")
    conf.message(f"Number of password imported: {len(paths_imported)}")
    if conf['convert']:
        conf.message("Forbidden chars converted")
        conf.message(f"Path separator used: {conf['separator']}")
    if conf['clean']:
        conf.message("Imported data cleaned")
    if conf['all']:
        conf.message("All data imported")
    for password, count in audit['breached']:
        conf.warning(f"Password breached {count} time(s): {password}")
    for password, details in audit['weak']:
        conf.warning(f"Weak password detected: {password} might be weak."
                     f" {zxcvbn_parse(details)}")
    for entry in audit['duplicated']:
        conf.warning(f"Duplicated passwords detected: "
                     f"{', '.join([item['path'] for item in entry])}")

    for paths, header in [
        (paths_imported, "Passwords imported"),
        (paths_exported, "Passwords exported")
    ]:
        if paths is not None:
            conf.message(header + ": " + str(len(paths)))
            paths.sort()
            for path in paths:
                conf.echo(path)


def main():
    """`pimport` and `pass import` common main."""
    conf = setup()
    decryptsource(conf)

    # Password managers detection
    cls_import = detectmanager(conf)
    cls_export = MANAGERS.get(conf['exporter'], cap=Cap.EXPORT)
    conf.verbose(f"Importing passwords from {cls_import.__name__} "
                 f"to {cls_export.__name__}")
    conf.verbose("Checking for breached passwords",
                 "on haveibeenpwned.com" if conf['pwned'] else '')

    # Import & export
    data = pass_import(conf, cls_import)
    paths_imported, paths_exported, audit = pass_export(conf, cls_export, data)

    # Success!
    report(conf, paths_imported, paths_exported, audit)


if __name__ == "__main__":
    main()



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import json
import re
from pass_import.core import register_managers, register_detecters
from pass_import.formats.csv import CSV
from pass_import.formats.json import JSON


class OnePasswordCSV(CSV):
    """Importer for 1password 6 in CSV format."""
    name = '1password'
    default = False
    version = '6'
    default = False
    url = 'https://1password.com'
    hexport = 'See this guide: https://support.1password.com/export'
    himport = 'pass import 1password file.csv'
    keys = {
        'title': 'Title',
        'password': 'Password',
        'login': 'Username',
        'url': 'URL',
        'comments': 'Notes',
        'group': 'Type'
    }


class OnePassword4CSV(CSV):
    """Importer for 1password 4 in CSV format."""
    name = '1password'
    default = False
    version = '4'
    only = True
    url = 'https://1password.com'
    hexport = 'See this guide: https://support.1password.com/export'
    himport = 'pass import 1password file.csv'
    keys = {
        'title': 'title',
        'password': 'password',
        'login': 'username',
        'url': 'url',
        'comments': 'notes'
    }


class OnePassword8CSV(CSV):
    """Importer for 1password 8 in CSV format."""
    name = '1password'
    version = '8'
    url = 'https://1password.com'
    hexport = 'See this guide: https://support.1password.com/export'
    himport = 'pass import 1password file.csv'
    keys = {
        'title': 'Title',
        'url': 'Url',
        'login': 'Username',
        'password': 'Password',
        'otpauth': 'OTPAuth',
        'favorite': 'Favorite',
        'archived': 'Archived',
        'tags': 'Tags',
        'comments': 'Notes'
    }


class OnePassword4PIF(JSON):
    """Importer for 1password 4 in PIF format.

    :param list ignore: List of key in the PIF file to not try to import.

    """
    name = '1password'
    format = '1pif'
    default = False
    version = '4'
    url = 'https://1password.com'
    hexport = 'See this guide: https://support.1password.com/export'
    himport = 'pass import 1password file.1pif'
    encoding = 'utf-8-sig'
    ignore = {'keyID', 'typeName', 'uuid', 'openContents', 'URLs'}
    keys = {
        'title': 'title',
        'password': 'password',
        'login': 'username',
        'url': 'location',
        'comments': 'notesPlain',
        'group': 'folderUuid',
        'tags': 'tags'
    }

    # Import methods

    @staticmethod
    def pif2json(file):
        """Convert 1pif to json: https://github.com/eblin/1passpwnedcheck."""
        data = file.read()
        cleaned = re.sub(r'(?m)^\*\*\*.*\*\*\*\s+', '', data)
        cleaned = cleaned.split('\n')
        # On 1Password v7.9.11 (macOS), 1PIF export produces 1 extra empty line
        cleaned = [v for v in cleaned if len(v) > 0]
        cleaned = ','.join(cleaned).rstrip(',')
        cleaned = f'[{cleaned}]'
        # JSON string with eventual special characters are encoded properly
        # eg: NUL, TAB
        cleaned = json.dumps(json.loads(cleaned, strict=False))
        return json.loads(cleaned)

    def parse(self):
        """Parse PIF based file."""
        jsons = self.pif2json(self.file)
        keys = self.invkeys()
        folders = {}
        for item in jsons:
            if item.get('typeName', '') == 'system.folder.Regular':
                key = item.get('uuid', '')
                folders[key] = {
                    'group': item.get('title', ''),
                    'parent': item.get('folderUuid', '')
                }

            elif item.get('typeName', '') == 'webforms.WebForm':
                if item.get('trashed', False):
                    continue
                entry = {}
                scontent = item.pop('secureContents', {})
                fields = scontent.pop('fields', [])
                for field in fields:
                    name = field.get('name', '')
                    designation = field.get('designation', '')
                    jsonkey = name or designation
                    key = keys.get(jsonkey, jsonkey)
                    entry[key] = field.get('value', '')

                sections = scontent.get('sections', [])
                for section in sections:
                    for field in section.get('fields', []):
                        value = field.get('v', '')
                        if value.startswith('otpauth://'):
                            entry['otpauth'] = value

                item.update(scontent)
                for key, value in item.items():
                    if key not in self.ignore:
                        entry[keys.get(key, key)] = value

                tags = []
                if 'openContents' in item:
                    open_contents = item['openContents']
                    tags = open_contents.get('tags', [])

                entry['tags'] = tags
                self.data.append(entry)
        self._sortgroup(folders)

    # Format recognition method

    def is_format(self):
        """Return True if the file is a 1PIF file."""
        try:
            self.jsons = self.pif2json(self.file)
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            return False
        return True

    def checkheader(self, header, only=False):
        """No header check is needed."""
        return True


register_managers(OnePassword8CSV,
                  OnePasswordCSV, OnePassword4CSV, OnePassword4PIF)
register_detecters(OnePassword4PIF)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import json

from pass_import.formats.json import JSON


class OTP(JSON):
    """Base class for OTP based importers."""
    content = ''

    # Import methods

    @staticmethod
    def _otp(item):
        otp = f"otpauth://{item.get('type', 'totp').lower()}/totp-secret?"
        otp += f"secret={item['secret']}&issuer={item['label']}"
        for setting in ['algorithm', 'digits', 'counter', 'period']:
            if setting in item:
                otp += f"&{setting}={item[setting]}"
        return otp

    def parse(self):
        """Parse OTP based file."""
        jsons = json.loads(self.content)
        for item in jsons:
            entry = {}
            entry['title'] = item['label']
            entry['otpauth'] = self._otp(item)

            for key in ['type', 'thumbnail', 'last_used']:
                entry[key] = str(item.get(key, '')).lower()
            entry['tags'] = ', '.join(item['tags'])
            self.data.append(entry)

    # Context manager method

    def open(self):
        """Parse OTP based file."""
        super().open()
        self.content = self.file.read()



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2022 Stefan Marsiske <sphinx@ctrlc.hu>.
# Copyright (C) 2022 Alexandre PUJOL <alexandre@pujol.io>.
#

import urllib

try:
    from pwdsphinx import sphinx
    PWDSPHINX = True
except ImportError:
    PWDSPHINX = False

from pass_import.core import register_managers
from pass_import.errors import PMError
from pass_import.manager import PasswordExporter
from pass_import.tools import getpassword


class Sphinx(PasswordExporter):
    """Exporter for Sphinx."""
    name = 'sphinx'
    default = True
    url = 'https://github.com/stef/pwdsphinx/'

    # Export methods

    def insert(self, entry):
        """Insert a password entry into Sphinx."""
        if entry.pop('is_a_history_entry', False):
            raise PMError(f"Skipping historical '{entry['path']}'")
        pwd = entry.pop('password', '')
        if not pwd:
            raise PMError(f"Skipping empty password '{entry['path']}'")
        user = entry.pop('login', '')
        if not user:
            raise PMError(f"Skipping '{entry['path']}' has no user")
        url = entry.pop('url', None)
        if url:
            url = urllib.parse.urlparse(url)
            host = f"{url.hostname}"
            if url.port:
                host = f"{host}:{url.port}"
        else:
            host = entry.pop('path', '')
        # print("user", user, "host", host, "pwd", pwd)

        try:
            sphinx.bin2pass.pass2bin(pwd, None)
        except OverflowError:
            raise PMError(f"Failed to import password for {user}@{host}"
                          " to sphinx: password too long.")

        s = sphinx.connect()
        try:
            sphinx.create(s, self.masterpassword, user, host, target=pwd)
        except Exception:
            raise PMError(f"Failed to export {user}@{host} to sphinx.")
        s.close()

    # Context manager methods

    def exist(self):
        """Nothing to do."""
        return True

    def open(self):
        """Get Sphinx master password."""
        if not PWDSPHINX:
            raise ImportError(name='pwdsphinx')
        self.masterpassword = getpassword("sphinx").encode('utf8')

    def close(self):
        """Forget Sphinx master password."""
        self.masterpassword = None


register_managers(Sphinx)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import csv
from typing import List

from pass_import.core import Cap, register_detecters
from pass_import.detecter import Formatter
from pass_import.errors import FormatError
from pass_import.manager import PasswordImporter


class CSV(Formatter, PasswordImporter):
    """Base class for CSV based importers.

    :param list fieldnames: The list of CSV field names
    :param str csv_header: If required special csv header to look for in the
        file.

    """
    cap = Cap.FORMAT | Cap.IMPORT
    format = 'csv'
    csv_header = ''
    fieldnames: List = []
    quotechar = '"'
    reader = None

    # Import method

    def parse(self):
        """Parse CSV based file."""
        fields = None if not self.fieldnames else self.fieldnames
        self.reader = csv.DictReader(self.file,
                                     fieldnames=fields,
                                     delimiter=self.delimiter,
                                     quotechar=self.quotechar)
        if not self.checkheader(self.header(), self.only):
            raise FormatError()

        keys = self.invkeys()
        for row in self.reader:
            entry = {}
            for col in row:
                entry[keys.get(col, col)] = row.get(col, None)

            self.data.append(entry)

    # Format recognition methods

    def is_format(self) -> bool:
        """Return True if the file is a CSV file."""
        try:
            dialect = csv.Sniffer().sniff(self.file.read(4096),
                                          delimiters=self.delimiter)
            if dialect.quotechar != self.quotechar:  # pragma: no cover
                return False
            self.file.seek(0)
            self.reader = csv.DictReader(self.file, dialect=dialect)

            # Context:
            #   1password can export data in:
            #   - 1PIF (json like format)
            #   - CSV
            # Problem:
            #   CSV sniffer considers the following line as a CSV
            #   line and will provide a dialect for it.
            #
            #   line: {"field0":"foo", ..., fieldX:"baz"}
            #
            #   This is clearly a JSON formatted line.
            # Solution:
            #    If the line looks like a JSON, then consider the
            #    file not a CSV file
            def is_json_key_value_format(value):
                values = value.split(':', 2)
                return len(values) > 1

            def is_json_open_end_bracket(keys):
                count = len(keys)
                if count > 0:
                    if (
                            keys[0].startswith('{') and
                            keys[-1].endswith('}')
                    ):
                        return False
                return False

            # Read first data line when available
            for index, row in enumerate(self.reader):
                if index == 0:
                    # skip header
                    continue

                keys = [k for k in row.keys() if k is not None]
                if all(map(is_json_key_value_format, keys)):
                    if is_json_open_end_bracket(keys):
                        return False

                break
        except (csv.Error, UnicodeDecodeError):
            return False
        return True

    def checkheader(self, header: List, only: bool = False) -> bool:
        """Ensure the file header is the same than the pm header."""
        try:
            if only and len(self.reader.fieldnames) != len(header):
                return False
            for csvkey in header:
                if csvkey not in self.reader.fieldnames:
                    return False
            return True
        except csv.Error:
            return False

    @classmethod
    def header(cls):
        """Header for CSV file."""
        return cls.keys.values()


register_detecters(CSV)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import json
import os

from pass_import.core import register_managers
from pass_import.errors import FormatError, PMError
from pass_import.formats.cli import CLI
from pass_import.formats.csv import CSV
from pass_import.tools import getpassword


class LastpassCLI(CLI):
    """Importer & Exporter for Lastpass using lpass.

    Binary attachments are not supported by Lastpass
    The Lastpass login is given either by the prefix or by the configuration
    file under the lastpass.login entry.

    Example:
    -------
    .. code-block:: yml

        lastpass:
          login: <your email addresss>

    """
    name = 'lastpass'
    command = 'lpass'
    url = 'https://www.lastpass.com'
    himport = 'pass import lastpass <login>'
    keys = {
        'title': 'name',
        'login': 'username',
        'url': 'url',
        'comments': 'note',
        'group': 'group'
    }

    def __init__(self, prefix=None, settings=None):
        self._opt = []
        self.sep = '\\'

        settings = {} if settings is None else settings
        conf = settings.get('lastpass', {})
        prefix = conf.get('login', prefix)

        super().__init__(prefix, settings)
        self._setenv('LPASS_HOME')
        self._setenv('LPASS_AUTO_SYNC_TIME')
        self._setenv('LPASS_AGENT_TIMEOUT')
        self._setenv('LPASS_AGENT_DISABLE')
        self._setenv('LPASS_PINENTRY')
        self._setenv('LPASS_DISABLE_PINENTRY', value='1')
        self._setenv('LPASS_ASKPASS')
        self._setenv('LPASS_CLIPBOARD_COMMAND')

    def _path(self, path, rep=os.sep):
        r"""Lpass is not consitent with / and '\\'. Replace them by os.sep."""
        return path.replace('/', rep).replace(self.sep, rep)

    def sync(self):
        """Force a synchronization of the local cache with the servers."""
        self._command(['sync', '--color=never'])

    def list(self, path=''):
        """List the paths in the password store repository.

        :param str path: Root path to the password repository to list.
        :return list: Return a list of unique ID in the store.

        """
        uids = []
        path = path.replace(self.sep, os.sep)

        arg = ['ls', '--format=%ai|%aN', '--sync=now', '--color=never']
        data = self._command(arg).split('\n')
        for line in data:
            if '|' in line:
                uid, group = line.split('|', 1)
                group = group.replace(self.sep, os.sep)
                if not group.endswith(os.sep) and path in group:
                    uids.append(uid)
        return uids

    # Import methods

    def show(self, uid):
        """Decrypt a lastpass entry and read the credentials.

        lpass do not show the same data with the --json option and without.
        To retrieve the full entry, both --json and --format option need to
        be used.

        :param str uid: UniqueID to the password entry to decrypt.
        :return dict: Return a dictionary with of the password entry.

        """
        entry = {}

        # lpass show --json
        ignores = {'fullname'}
        keys = self.invkeys()
        jsons = self._command(['show', '--json', uid])
        item = json.loads(jsons).pop()
        for key, value in item.items():
            if key not in ignores:
                entry[keys.get(key, key)] = value
        entry['group'] = self._path(item['group'])

        # lpass show --format
        ignores = {'Username', 'Password', 'URL'}
        arg = ['show', '--color=never',
               "--format=%fn|%fv", '--color=never', uid]
        data = self._command(arg).split('\n')
        data.pop()
        data.pop(0)
        for line in data:
            if '|' in line:
                key, value = line.split('|', 1)
                if key not in ignores:
                    entry[key] = value

        # Special cleanup
        if entry.get('url', '') == 'http://':
            entry['url'] = ''
        return entry

    def parse(self):
        """Parse Lastpass repository using lpass."""
        uniqueids = self.list(self.root)
        if not uniqueids:
            raise FormatError('empty password store.')

        for uniqueid in uniqueids:
            entry = self.show(uniqueid)
            self.data.append(entry)

    # Export methods

    def remove(self, uid):
        """Move an entry to the Trash."""
        arg = ['rm', '--color=never', uid]
        self._command(arg)

    def insert(self, entry):
        """Insert a password entry into lastpass using lpass."""
        path = os.path.join(self.root, entry['path'])
        entry['group'] = os.path.dirname(path)
        entry['title'] = os.path.basename(path)
        path = entry['group'].replace(os.sep, self.sep) + '/' + entry['title']

        # Remove entries with the same name.
        uids = self.list(path)
        if uids:
            if not self.force:
                raise PMError(f"An entry already exists for {path}.")
            for uid in uids:
                self.remove(uid)

        # Insert the entry into lastpass
        seen = {'path', 'title', 'group'}
        exportkeys = {
            'title': 'Name',
            'password': 'Password',
            'login': 'Username',
            'url': 'URL',
            'comments': 'Notes',
        }
        data = ''
        for key in self.keyslist:
            if key in seen:
                continue
            if key in entry:
                data += f"{exportkeys.get(key, key)}: {entry[key]}\n"
            seen.add(key)

        if self.all:
            for key, value in entry.items():
                if key in seen:
                    continue
                data += f"{key}: {value}\n"

        arg = ['add', '--sync=now', '--non-interactive', '--color=never', path]
        self._command(arg, data)

    # Context manager methods

    def open(self):
        """Sign in to your Lastpass account."""
        if self.prefix == '':
            raise PMError("Your Lastpass username is empty")

        status = [self._binary, 'status', '--quiet']
        res, _, _ = self._call(status)
        if res:
            login = ['login', '--trust', '--color=never', self.prefix]
            password = getpassword('Lastpass')
            res = self._command(login, password)

    def close(self):
        """Synchronise and sign out of your Lastpass account."""
        self.sync()


class LastpassCSV(CSV):
    """Importer for Lastpass in CSV format."""
    name = 'lastpass'
    default = False
    url = 'https://www.lastpass.com'
    hexport = 'More Options > Advanced > Export'
    keys = {
        'title': 'name',
        'password': 'password',
        'login': 'username',
        'url': 'url',
        'comments': 'extra',
        'group': 'grouping'
    }

    def parse(self):
        """Parse Lastpass CSV file."""
        super().parse()
        for entry in self.data:
            if 'group' in entry and entry['group'] is None:
                # LastPass will truncate everything after `$` in a
                # secure note entry when exporting as a CSV, including
                # any closing ", leaving the file in a corrupt
                # state. Triggering this is likely a symptom of such a
                # corrupted export.
                #
                # Likewise, it also has problems exporting single
                # quotes in the password field, causing all data prior
                # to the single quote (including the url field, etc.)
                # to be truncated, leading to the parser thinking the
                # path field wasn't included, and incorrectly
                # resulting in a value of None.
                raise FormatError(f'Invalid group in entry:\n{entry}.')
            entry['group'] = entry.get('group', '').replace('\\', os.sep)


register_managers(LastpassCLI, LastpassCSV)



# -*- encoding: utf-8 -*-
# pass import - Passwords importer swiss army knife
# Copyright (C) 2017-2024 Alexandre PUJOL <alexandre@pujol.io>.
#

import json

from pass_import.core import Cap, register_detecters
from pass_import.detecter import Formatter
from pass_import.manager import PasswordImporter


class JSON(Formatter, PasswordImporter):
    """Base class for JSON based importers."""
    cap = Cap.FORMAT | Cap.IMPORT
    format = 'json'
    json_header = {}
    jsons = None

    # Import methods

    def parse(self):
        """Parse JSON based file."""
        raise NotImplementedError()

    # Format recognition methods

    def is_format(self) -> bool:
        """Return True if the file is a JSON file."""
        try:
            self.jsons = json.loads(self.file.read())
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            return False
        return True

    def _ensureheader(self, data, header) -> bool:
        """Ensure the data is in the header format.

        :return bool:
            True or False either the header match the json data.

        """
        # pylint: disable=too-many-return-statements
        if isinstance(header, type):
            if data is not None and not isinstance(data, header):
                return False
            return True

        if isinstance(data, list):
            if not isinstance(header, list):
                return False
            for item in data:
                if not self._ensureheader(item, header[0]):
                    return False
            return True

        if isinstance(data, dict):
            if not isinstance(header, dict):
                return False
            for key, value in header.items():
                if key not in data or not self._ensureheader(data[key], value):
                    return False
            return True

        return data == header

    def checkheader(self, header, only=False) -> bool:
        """Ensure the file header is the same than the pm header."""
        return self._ensureheader(self.jsons, header)

    @classmethod
    def header(cls):
        """Header for JSON file."""
        return cls.json_header


register_detecters(JSON)
